# Copyright 2023, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from collections import defaultdict
from collections.abc import Callable
from enum import Enum
from typing import Any

import cocotb
from cocotb.triggers import Event


class EventEmitter:
    """Core support for publishing events and subscribing to them"""

    def __init__(self) -> None:
        self._handlers = defaultdict(list)
        self._ready = Event()
        self._waiting = defaultdict(list)

    def subscribe(self, event: Enum, callback: Callable) -> None:
        """
        Subscribe to an event being published by this component.

        :param event:    Enumerated event
        :param callback: Method to call when the event occurs, this must accept
                         arguments of component, event type, and an associated
                         object
        """
        if not isinstance(event, Enum):
            raise TypeError(f"Event should inherit from Enum, unlike {event}")
        self._handlers[event].append(callback)

    def subscribe_all(self, callback: Callable) -> None:
        """
        Subscribe to all events published by this component.

        :param callback: Method to call when the event occurs, this must accept
                         arguments of component, event type, and an associated
                         object
        """
        self._handlers["*"].append(callback)

    def unsubscribe_all(self, event: Enum | None = None) -> None:
        """
        De-register all subscribers for a given event (when event is not None),
        or all subscribers from all events (when event is None).

        :param event: Optional enumerated event to unsubscribe, or None to
                      unsubscribe all subscribers for all events
        """
        if event is None:
            self._handlers.clear()
            self._waiting.clear()
        elif not isinstance(event, Enum):
            raise TypeError(f"Event should inherit from Enum, unlike {event}")
        else:
            self._handlers[event].clear()
            self._waiting[event].clear()

    def publish(self, event: Enum, obj: Any) -> None:
        """
        Publish an event and deliver it to any registered subscribers.

        :param event: Enumerated event
        :param obj:   Object associated to the event
        """
        # Call direct handlers
        for handler in self._handlers["*"] + self._handlers[event]:
            call = handler(self, event, obj)
            if asyncio.iscoroutine(call):
                cocotb.start_soon(call)
        # Trigger pending events
        events = self._waiting[event][:]
        self._waiting[event].clear()
        for event in events:
            event.set(data=obj)

    def _get_wait_event(self, event: Enum) -> Event:
        """
        Internal method for generating cocotb Events tied to a specific enumerated
        event trigger.

        :param event: Enumerated event trigger
        :returns:     The registered cocotb Event
        """
        evt = Event()
        self._waiting[event].append(evt)
        return evt

    async def wait_for(self, event: Enum) -> Any:
        """
        Wait for a specific enumerated event to occur and return the data that
        was associated to it.

        :param event: Enumerated event trigger
        :returns:     Data associated with the event
        """
        evt = self._get_wait_event(event)
        await evt.wait()
        return evt.data



# Copyright 2023, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from collections import defaultdict
from collections.abc import Callable
from enum import Enum, auto
from typing import Any

import cocotb
from cocotb.log import SimLog
from cocotb.triggers import Event, First, Lock, RisingEdge, Timer
from cocotb.utils import get_sim_time

from .monitor import BaseMonitor, MonitorEvent
from .transaction import BaseTransaction


class QueueEmptyError(Exception):
    pass


class ChannelTimeoutError(AssertionError):
    pass


class WindowResidenceError(AssertionError):
    pass


class DrainPolicy(Enum):
    """Specify the draining behaviour for a scoreboard channel"""

    MON_AND_REF = auto()
    """Block until both monitor and reference queues are empty"""
    MON_ONLY = auto()
    """Block until only the monitor queue is empty"""
    REF_ONLY = auto()
    """Block until only the reference queue is empty"""
    NON_BLOCKING = auto()
    """Do not wait for either queue to drain"""


class Queue:
    """
    A custom queue implementation that allows peeking onto the head of the queue,
    which assists with the implementation of funnel-type channels.
    """

    def __init__(self) -> None:
        self._entries = []
        self._on_push = None

    def __len__(self) -> int:
        return self.level

    def __getitem__(self, key) -> BaseTransaction | list[BaseTransaction]:
        return self._entries[key]

    @property
    def level(self) -> int:
        return len(self._entries)

    @property
    def on_push_event(self) -> Event:
        """Expose the event that will be fired on the next push"""
        if self._on_push is None:
            self._on_push = Event()
        return self._on_push

    def push(self, data: Any) -> None:
        """
        Push an entry into the queue, notifying any observers that have
        registered an 'on-push' event.

        :param data: Data to push into the queue
        """
        self._entries.append(data)
        if self._on_push is not None:
            self._on_push.set()
        self._on_push = None

    async def pop(self, index: int = 0) -> Any:
        """
        Pop an entry from the queue, if necessary blocking until one is available.

        :param index: Index of the item to pop
        :returns:     Object from the head of the queue
        """
        if len(self._entries) == 0:
            await self.wait()
        return self._entries.pop(index)

    def wait(self) -> None:
        """Register an 'on-push' event and wait for it to be set by a push"""
        return self.on_push_event.wait()

    async def wait_for_not_empty(self) -> None:
        """Wait until at least one entry exists in the queue"""
        if self.level == 0:
            await self.wait()

    def peek(self) -> Any:
        if len(self._entries) == 0:
            raise QueueEmptyError()
        return self._entries[0]


class Channel:
    """
    A channel gathers transactions from a monitor and a reference model of some
    form. When both queues contain an entry, the top-most entry is popped from
    both queues and tested for equivalence. Any mismatches are reported to the
    scoreboard.

    :param name:         Name of this scoreboard channel
    :param monitor:      Handle to the monitor capturing traffic from the DUT
    :param log:          Handle to the scoreboard log
    :param filter_fn:    Function to filter or modify captured transactions
    :param timeout_ns:   Optional timeout to allow for a object sat at the front
                         of the monitor queue to remain unmatched (in nanoseconds,
                         a value of None disables the timeout mechanism)
    :param polling_ns:   How frequently to poll to check for unmatched items stuck
                         in the monitor queue in nanoseconds (defaults to 100 ns)
    :param match_window: Where precise ordering of expected transactions is not
                         known, a positive integer matching window can be used
                         to match any of the next N transactions in the reference
                         queue (where N is set by match_window)
    :param drain_policy: Specify how the handle should handle draining and which
                         queues should be considered when blocking testcase
                         shutdown
    """

    def __init__(
        self,
        name: str,
        monitor: BaseMonitor,
        log: SimLog,
        filter_fn: Callable | None,
        timeout_ns: int | None = None,
        polling_ns: int = 100,
        drain_policy: DrainPolicy = DrainPolicy.MON_AND_REF,
        match_window: int = 1,
    ) -> None:
        self.name = name
        self.monitor = monitor
        self.log = log
        self.filter_fn = filter_fn
        self.timeout_ns = timeout_ns
        self.polling_ns = polling_ns
        self.drain_policy = drain_policy
        assert self.drain_policy in list(
            DrainPolicy
        ), f"Unsupported draining mode {self.drain_policy}"
        self.match_window = match_window or 1
        assert (
            isinstance(self.match_window, int) and self.match_window > 0
        ), "Channel matching window must be a positive integer"
        self._q_mon = Queue()
        self._q_ref = Queue()
        self._lock = Lock()
        self._matched = 0
        self._mismatched = 0
        self._residence = defaultdict(lambda: 0)

        def _sample(mon: BaseMonitor, evt: MonitorEvent, obj: BaseTransaction) -> None:
            if mon is self.monitor and evt is MonitorEvent.CAPTURE:
                # If a filter function was provided, apply it
                if self.filter_fn is not None:
                    obj = self.filter_fn(mon, evt, obj)
                # A filter can drop the transaction, so test for None
                if obj is not None:
                    self.push_monitor(obj)

        self.monitor.subscribe(MonitorEvent.CAPTURE, _sample)
        cocotb.start_soon(self._polling())

    @property
    def monitor_depth(self) -> int:
        """Number of captured packets queued up by the DUT monitor"""
        return self._q_mon.level

    @property
    def reference_depth(self) -> int:
        """Number of expected packets queued up by the reference model"""
        return self._q_ref.level

    @property
    def matched(self) -> int:
        """Number of matched packets between monitor and reference model"""
        return self._matched

    @property
    def mismatched(self) -> int:
        """Number of mismatched packets between monitor and reference model"""
        return self._mismatched

    @property
    def total(self) -> int:
        """Total number of packets matched/mismatched"""
        return self._matched + self._mismatched

    def push_monitor(self, *transactions: BaseTransaction) -> None:
        """
        Push one or more captured transactions into the monitor's queue.

        :param *transactions: Captured transactions
        """
        for transaction in transactions:
            assert isinstance(transaction, BaseTransaction)
            self._q_mon.push(transaction)

    def push_reference(self, *transactions: BaseException) -> None:
        """
        Push one or more captured transactions into the reference (model) queue.

        :param *transactions: Captured transactions
        """
        for transaction in transactions:
            assert isinstance(transaction, BaseTransaction)
            self._q_ref.push(transaction)

    async def _dequeue(self) -> tuple[BaseTransaction, BaseTransaction]:
        """
        Dequeue the top-most transaction from both the monitor and reference
        queues.

        :returns: Tuple of the monitor transaction and reference transaction
        """
        # Wait for monitor to capture a transaction
        await self._q_mon.wait_for_not_empty()
        # When a monitor transaction arrives, lock out the drain procedure
        await self._lock.acquire()
        # If matching window is 1, wait for the first reference transaction
        if self.match_window == 1:
            next_ref = await self._q_ref.pop()
        # If matching window > 1,
        else:
            # Peek at the captured transaction and search the next N transactions
            peek_mon = self._q_mon[0]
            next_ref = None
            while next_ref is None:
                # Search within the match window
                for idx, peek_ref in enumerate(self._q_ref[: self.match_window]):
                    # Track how many cycles this item has been present in the matching window
                    self._residence[id(peek_ref)] += 1
                    # Check if this item has been in the matching window too long?
                    if self._residence[id(peek_ref)] > self.match_window:
                        raise WindowResidenceError(
                            f"Item has been present in the matching window of scoreboard "
                            f"channel {self.name} for longer than the window length of "
                            f"{self.match_window}: {peek_ref}"
                        )
                    # If the reference and monitor transactions match...
                    if peek_ref == peek_mon:
                        # ...pop the reference transaction and...
                        next_ref = await self._q_ref.pop(idx)
                        # ...stop tracking its residence time
                        del self._residence[id(peek_ref)]
                        break
                # If no match found...
                else:
                    # ...if queue length less than the match window, wait for a push
                    if len(self._q_ref) < self.match_window:
                        await self._q_ref.on_push_event.wait()
                    # ...otherwise pop the zeroeth entry
                    else:
                        next_ref = await self._q_ref.pop(0)
        # Pop the front of the monitor queue
        next_mon = await self._q_mon.pop()
        # Return monitor-reference pair (don't release lock yet)
        return next_mon, next_ref

    async def _polling(self) -> None:
        """
        Polling loop that checks for items getting stuck in the channel's
        monitor queue
        """
        while True:
            # Wait for polling delay
            await Timer(self.polling_ns, units="ns")
            # Check for object at the front of the monitor queue
            if (
                (self.timeout_ns is not None)
                and (self._q_mon.level > 0)
                and (
                    (age := (get_sim_time(units="ns") - self._q_mon.peek().timestamp))
                    > self.timeout_ns
                )
            ):
                self.log.error(
                    f"Object at the front of the of the {self.name} monitor "
                    f"queue has been stalled for {age} ns which exceeds the "
                    f"configured timeout of {self.timeout_ns} ns"
                )
                self.report()
                raise ChannelTimeoutError(f"Channel {self.name} timed out")

    async def loop(self, mismatch: Callable, match: Callable | None = None) -> None:
        """
        Continuously dequeue pairs of transactions from the monitor and reference
        queues and report mismatches to the scoreboard via the callback.

        :param mismatch: Callback to execute whenever a mismatch occurs, providing
                         arguments of the channel (self), the monitor object, and
                         the reference object
        """
        while True:
            # Wait for a monitor-reference pair to be dequeued
            mon, ref = await self._dequeue()
            if mon != ref:
                self._mismatched += 1
                mismatch(self, mon, ref)
            else:
                self._matched += 1
                if match is not None:
                    match(self, mon, ref)
            # Release the lock after the comparison completes
            self._lock.release()

    async def drain(self) -> None:
        """Block until the channel's monitor and reference queues empty"""
        # Start draining
        match self.drain_policy:
            case DrainPolicy.MON_AND_REF:
                while self.monitor_depth > 0 or self.reference_depth > 0:
                    await RisingEdge(self.monitor.clk)
            case DrainPolicy.MON_ONLY:
                while self.monitor_depth > 0:
                    await RisingEdge(self.monitor.clk)
            case DrainPolicy.REF_ONLY:
                while self.reference_depth > 0:
                    await RisingEdge(self.monitor.clk)
        # Helper messages to flag if reference or monitor queues contain entries
        if self.reference_depth > 0:
            self.log.warning(
                f"Reference queue of scoreboard channel {self.name} still contains "
                f"{self.reference_depth} items but is ignored by the draining mode"
            )
        if self.monitor_depth > 0:
            self.log.warning(
                f"Monitor queue of scoreboard channel {self.name} still contains "
                f"{self.monitor_depth} items but is ignored by the draining mode"
            )
        # Wait for the lock to ensure a comparison is not still underway
        await self._lock.acquire()
        # Release the lock (in case we call drain multiple times)
        self._lock.release()

    def report(self) -> None:
        """
        Report the status of this channel detailing number of entries in the
        monitor and reference queues, and the queued transactions at the head
        of those queues
        """
        # Report matches/mismatches
        self.log.info(
            f"Channel {self.name} paired {self.total} transactions of which "
            f"{self.mismatched} mismatches were detected"
        )
        # Report outstanding transactions
        mon_depth = self.monitor_depth
        ref_depth = self.reference_depth
        if mon_depth > 0 or ref_depth > 0:
            self.log.error(
                f"Channel {self.name} has {mon_depth} entries captured from "
                f"monitor and {ref_depth} entries queued from reference model"
            )
        if mon_depth > 0:
            self.log.info(f"Packet at head of {self.name}'s monitor queue:")
            self.log.info(self._q_mon.peek().tabulate())
        if ref_depth > 0:
            self.log.info(f"Packet at head of {self.name}'s reference queue:")
            self.log.info(self._q_ref.peek().tabulate())


class FunnelChannel(Channel):
    """
    An extended scoreboard channel where the order of data exiting the monitor
    is not strictly defined, often due to the hardware interleaving different
    streams. Reference data can be pushed into one or more named queues and when
    monitor packets arrive any queue head is valid.

    :param name:         Name of this scoreboard channel
    :param monitor:      Handle to the monitor capturing traffic from the DUT
    :param log:          Handle to the scoreboard log
    :param filter_fn:    Function to filter or modify captured transactions
    :param ref_queues:   List of reference queue names
    :param timeout_ns:   Optional timeout to allow for a object sat at the front
                         of the monitor queue to remain unmatched (in nanoseconds,
                         a value of None disables the timeout mechanism)
    :param polling_ns:   How frequently to poll to check for unmatched items stuck
                         in the monitor queue in nanoseconds (defaults to 100 ns)
    :param drain_policy: Specify how the handle should handle draining and which
                         queues should be considered when blocking testcase
                         shutdown
    """

    def __init__(
        self,
        name: str,
        monitor: BaseMonitor,
        log: SimLog,
        filter_fn: Callable | None,
        ref_queues: list[str] | tuple[str],
        timeout_ns: int | None = None,
        polling_ns: int = 100,
        drain_policy: DrainPolicy = DrainPolicy.MON_AND_REF,
    ) -> None:
        super().__init__(
            name,
            monitor,
            log,
            filter_fn,
            timeout_ns=timeout_ns,
            polling_ns=polling_ns,
            drain_policy=drain_policy,
        )
        self._q_ref = {x: Queue() for x in ref_queues}

    @property
    def reference_depth(self) -> int:
        return sum(x.level for x in self._q_ref.values())

    def push_reference(self, queue: str, *transactions: BaseException) -> None:
        """
        Push one or more captured transactions into a given reference (model)
        queue.

        :param queue:         Name of the queue to push into
        :param *transactions: Captured transactions
        """
        assert isinstance(queue, str) and queue in self._q_ref
        for transaction in transactions:
            assert isinstance(transaction, BaseTransaction)
            self._q_ref[queue].push(transaction)

    async def _dequeue(self) -> tuple[BaseTransaction, BaseTransaction]:
        """
        Dequeue the top-most transaction from both the monitor and reference
        queues, searching through the reference queues for a matching object to
        the monitor's captured transaction.

        :returns: Tuple of the monitor transaction and reference transaction
        """
        # Wait for monitor to capture a transaction
        await self._q_mon.wait_for_not_empty()
        # Once a monitor transaction arrives, lock out the drain procedure
        await self._lock.acquire()
        # Peek at the front of all of the queues
        while True:
            next_mon = self._q_mon.peek()
            any_empty = False
            for queue in self._q_ref.values():
                any_empty = any_empty or (queue.level == 0)
                if queue.level > 0 and queue.peek() == next_mon:
                    await self._q_mon.pop()
                    next_ref = await queue.pop()
                    return next_mon, next_ref
            # If all queues contain objects but none matched, this is a mismatch!
            # NOTE: This will just pop the final queue in order to report miscompare
            if not any_empty:
                await self._q_mon.pop()
                next_ref = await queue.pop()
                return next_mon, next_ref
            # Wait for a reference object to be pushed to any queue
            await First(*(x.on_push_event.wait() for x in self._q_ref.values()))

    def report(self) -> None:
        """
        Report the status of this channel detailing number of entries in the
        monitor and reference queues, and the queued transactions at the head
        of those queues
        """
        # Report matches/mismatches
        self.log.info(
            f"Channel {self.name} paired {self.total} transactions of which "
            f"{self.mismatched} mismatches were detected"
        )
        # Report outstanding transactions
        mon_depth = self.monitor_depth
        ref_depth = self.reference_depth
        if mon_depth > 0 or ref_depth > 0:
            self.log.error(
                f"Channel {self.name} has {mon_depth} entries captured from "
                f"monitor and {ref_depth} entries queued from reference model"
            )
        if mon_depth > 0:
            self.log.info(f"Packet at head of {self.name}'s monitor queue:")
            self.log.info(self._q_mon.peek().tabulate())
        for key, queue in self._q_ref.items():
            if queue.level > 0:
                self.log.info(f"Packet at head of {self.name}'s reference queue '{key}':")
                self.log.info(queue.peek().tabulate())


class MiscompareError(Exception):
    """
    Raises a miscomparison as an exception with associated data.

    :param channel:   Channel that the miscompare occurred on
    :param monitor:   Transaction captured by the monitor
    :param reference: Expected transaction queued by the model
    """

    def __init__(
        self, channel: Channel, monitor: BaseTransaction, reference: BaseTransaction
    ) -> None:
        super().__init__()
        self.channel = channel
        self.monitor = monitor
        self.reference = reference


class Scoreboard:
    """
    Scoreboard for comparing captured and reference transactions. The scoreboard
    is divided into different 'channels', one per monitor, that handle the
    comparison of captured and reference objects and reporting of any differences.

    :param tb:         Reference to the testbench
    :param fail_fast:  Stop the test as soon as a mismatch is reported (default: False)
    :param postmortem: Enter debug on a mismatch (default: False)
    """

    def __init__(
        self,
        tb,
        fail_fast: bool = False,
        postmortem: bool = False,
    ):
        self.fail_fast = fail_fast
        self.postmortem = postmortem
        self._mismatches = []
        self.tb = tb
        self.log = self.tb.fork_log("scoreboard")
        self.channels: dict[str, Channel] = {}

    def attach(
        self,
        monitor: BaseMonitor,
        filter_fn: Callable | None = None,
        queues: list[str] | tuple[str] | None = None,
        timeout_ns: int | None = None,
        polling_ns: int = 100,
        drain_policy: DrainPolicy = DrainPolicy.MON_AND_REF,
        match_window: int = 1,
    ) -> None:
        """
        Attach a monitor to the scoreboard, creating and scheduling a new
        channel in the process.

        :param monitor:      The monitor to attach
        :param queues:       List of reference queue names, this causes a funnel
                             type scoreboard channel to be used
        :param filter_fn:    A filter function that can either drop or modify items
                             recorded by the monitor prior to scoreboarding
        :param timeout_ns:   Optional timeout to allow for a object sat at the front
                             of the monitor queue to remain unmatched (in nanoseconds,
                             a value of None disables the timeout mechanism)
        :param polling_ns:   How frequently to poll to check for unmatched items stuck
                             in the monitor queue in nanoseconds (defaults to 100 ns)
        :param drain_policy: Specify how the handle should handle draining and
                             which queues should be considered when blocking
                             testcase shutdown
        :param match_window: Where precise ordering of expected transactions is not
                             known, a positive integer matching window can be used
                             to match any of the next N transactions in the reference
                             queue (where N is set by match_window)
        """
        assert monitor.name not in self.channels, f"Monitor known for '{monitor.name}'"
        if isinstance(queues, list | tuple) and len(queues) > 0:
            channel = FunnelChannel(
                monitor.name,
                monitor,
                self.tb.fork_log("scoreboard", "channel", monitor.name),
                filter_fn,
                queues,
                timeout_ns=timeout_ns,
                polling_ns=polling_ns,
                drain_policy=drain_policy,
            )
        else:
            channel = Channel(
                monitor.name,
                monitor,
                self.tb.fork_log("scoreboard", "channel", monitor.name),
                filter_fn,
                timeout_ns=timeout_ns,
                polling_ns=polling_ns,
                drain_policy=drain_policy,
                match_window=match_window,
            )
        self.channels[channel.name] = channel
        if self.log.getEffectiveLevel() <= logging.DEBUG:
            cocotb.start_soon(channel.loop(self._mismatch, self._match))
        else:
            cocotb.start_soon(channel.loop(self._mismatch))

    async def drain(self) -> None:
        """Block until all chains of the scoreboard have been drained"""
        for name, channel in self.channels.items():
            self.log.info(f"Draining scoreboard channel '{name}'...")
            await channel.drain()
            self.log.info(f"Drained scoreboard channel '{name}'")

    def _mismatch(
        self, channel: Channel, monitor: BaseTransaction, reference: BaseTransaction
    ) -> None:
        """
        Callback whenever a channel detects a mismatch between captured and
        reference objects. If fail_fast is enabled, then an assertion will fire.

        :param channel:   The channel reporting the mismatch
        :param monitor:   The transaction captured by a monitor
        :param reference: The reference transaction produced by a model
        """
        self.log.error(
            f"Mismatch on channel {channel.monitor.name} for transaction index "
            f"{channel.total-1}"
        )
        self.log.info(monitor.tabulate(reference))
        self._mismatches.append((channel, monitor, reference))
        if self.fail_fast:
            if self.postmortem:
                self.log.warning(
                    "Entering fast-fail postmortem - the 'monitor' variable gives "
                    "the captured transaction, 'reference' the reference "
                    "transaction, and 'channel' gives access to the scoreboard "
                    "channel"
                )
                breakpoint()
            raise MiscompareError(channel, monitor, reference)

    def _match(
        self, channel: Channel, monitor: BaseTransaction, reference: BaseTransaction
    ) -> None:
        """
        Callback whenever a channel detects a match between captured and
        reference objects, only called if monitor attached with DEBUG verbosity.

        :param channel:   The channel reporting the mismatch
        :param monitor:   The transaction captured by a monitor
        :param reference: The reference transaction produced by a model
        """
        self.log.debug(
            f"Match on channel {channel.monitor.name} for transaction index {channel.total-1}"
        )
        self.log.debug(monitor.tabulate(reference))

    @property
    def result(self) -> bool:
        """If no mismatches recorded return True, else return False"""
        return len(self._mismatches) == 0



# Copyright 2023, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import dataclasses
from enum import Enum
from typing import Any, Optional

from cocotb.triggers import Event
from cocotb.utils import get_sim_time
from tabulate import tabulate


@dataclasses.dataclass(kw_only=True)
class BaseTransaction:
    """
    Base transaction object type

    :param timestamp: The time at which the event occurred
    :param _f_event:  Forastero event type to trigger on
    :param _c_event:  The cocotb Event to trigger when _f_event is reached
    """

    timestamp: int = dataclasses.field(
        default_factory=lambda: get_sim_time(units="ns"), compare=False
    )

    _f_event: Enum | None = dataclasses.field(default=None, compare=False)
    _c_event: Event | None = dataclasses.field(default=None, compare=False)

    def copy(self) -> "BaseTransaction":
        return copy.copy(self)

    def format(self, field: str, value: Any) -> str:
        """
        Subclasses of BaseTransaction may override this to format different
        fields however they prefer.

        :param field: Name of the field to format
        :param value: Value to format
        :returns: A string of the formatted value
        """
        if field == "timestamp":
            return f"{value} ns"
        elif isinstance(value, Enum):
            return f"{value.name} ({int(value)})"
        elif isinstance(value, int):
            return f"0x{value:X}"
        elif hasattr(value, "__int__"):
            return f"0x{int(value):X}\n({value})"
        else:
            return str(value)

    def tabulate(self, other: Optional["BaseTransaction"] = None) -> str:
        """
        Tabulate the fields of this transaction, optionally comparing against
        the fields of another object of this same type. When comparison is
        performed, mismatches are flagged in a separate column.

        :param other: Optional transaction to compare against
        :returns: String of tabulated data
        """
        # Collect table headers
        headers = ["Field"]
        if other is None:
            headers.append("Data")
        else:
            headers += ["Captured", "Expected", "Compared?", "Match?"]
        # Assemble rows
        rows = []
        for field in dataclasses.fields(self):
            if field.name in ("_f_event", "_c_event"):
                continue
            a_val = self.format(field.name, getattr(self, field.name))
            cols = [field.name, a_val]
            if other is not None:
                b_val = self.format(field.name, getattr(other, field.name))
                flag = ["", "!!!"][field.compare and (a_val != b_val)]
                cols += [b_val, ["no", "yes"][field.compare], flag]
            rows.append(cols)
        # Render the table
        return tabulate(rows, headers=headers, tablefmt="grid")



# Copyright 2023, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import json
import logging
import os
import random
import traceback
from collections import defaultdict
from collections.abc import Callable, Coroutine
from pathlib import Path
from typing import Any, ClassVar

import cocotb
from cocotb.clock import Clock
from cocotb.handle import HierarchyObject, ModifiableObject
from cocotb.log import SimLog, SimLogFormatter, SimTimeContextFilter
from cocotb.result import SimTimeoutError
from cocotb.task import Task
from cocotb.triggers import ClockCycles, Event, with_timeout

from .component import Component
from .driver import BaseDriver
from .io import IORole
from .monitor import BaseMonitor
from .scoreboard import DrainPolicy, Scoreboard
from .sequence import BaseSequence, SeqArbiter


def strtobool(val: str) -> bool:
    lval = val.strip().lower()
    if lval in (truthy := ("y", "yes", "t", "true", "on", "1")):
        return True
    elif lval in (falsey := ("n", "no", "f", "false", "off", "0")):
        return False
    else:
        raise ValueError(f"Invalid bool `{val}`, specify using one of `{truthy}` or `{falsey}`")


class BaseBench:
    """
    Base class for a Forastero testbench

    :param dut:        Handle to the DUT, provided by cocotb
    :param clk:        Handle to the primary clock signal
    :param rst:        Handle to the primary reset signal
    :param clk_drive:  Whether the primary clock signal should be driven
    :param clk_period: Tick period for the primary clock
    :param clk_units:  Units of the primary clock's period
    """

    TEST_REQ_PARAMS: ClassVar[dict[Any, list[tuple[str, Callable[[str], Any]]]]] = defaultdict(list)
    PARAM_FILE_PATH: ClassVar[str] = os.environ.get("TEST_PARAMS", None)
    PARAM_DEFAULTS: ClassVar[dict[str, Any]] = {
        # Random seed
        "seed": 0,
        # Hierarchical logging control
        # NOTE: 'tb' is the log hierarchy root
        "verbosity": {
            "tb": "info",
        },
        # Enable profiling by providing a path
        "profiling": None,
        # Enable fast failure
        "fail_fast": (os.environ.get("FAIL_FAST", "no").lower() == "yes"),
        # Postmortem (activate a Python breakpoint after a failure)
        "postmortem": False,
        # Testcase parameters
        "testcases": {},
    }

    def __init__(
        self,
        dut: HierarchyObject,
        clk: ModifiableObject | None = None,
        rst: ModifiableObject | None = None,
        clk_drive: bool = True,
        clk_period: float = 1,
        clk_units: str = "ns",
    ) -> None:
        # Hold a pointer to the DUT
        self.dut = dut
        # Promote clock & reset
        self.clk = clk
        self.rst = rst
        # Clock driving
        self.clk_drive = clk_drive
        self.clk_period = clk_period
        self.clk_units = clk_units
        # Alias logging methods
        self.log = SimLog("tb")
        self.debug = self.log.debug
        self.info = self.log.info
        self.warning = self.log.warning
        self.error = self.log.error
        # Tee log into a file with timestamping
        log_fh = logging.FileHandler(Path.cwd() / "sim.log")
        log_fh.addFilter(SimTimeContextFilter())
        log_fh.setFormatter(SimLogFormatter())
        self.log.addHandler(log_fh)
        # Set verbosity at all specified hierarchy levels
        for hierarchy, verbosity in self.get_parameter("verbosity").items():
            logging.getLogger(hierarchy).setLevel(getattr(logging, verbosity.upper()))
        # Create a child log for orchestration
        # NOTE: This should really only be used by internal testbench processes
        self._orch_log = self.fork_log("int", "orch")
        # Create a scoreboard
        self.scoreboard = Scoreboard(
            tb=self,
            fail_fast=self.get_parameter("fail_fast", False),
            postmortem=self.get_parameter("postmortem", False),
        )
        # Track components
        self._components = {}
        self._processes = {}
        self._teardown = []
        # Random seeding
        self.seed = int(self.get_parameter("seed", 0))
        self.info(f"Bench initialised with random seed {self.seed}")
        self.random = random.Random(self.seed)
        # Sequence handling
        self._arbiter = SeqArbiter(self.fork_log("int", "seq"), self.random)
        self._sequences = []
        # Events
        self.evt_ready = Event()

    def fork_log(self, *scope: str) -> SimLog:
        """
        Create a new descedent of the root simulation log with a given context.

        :param *scope: A particular scope as a list of strings
        """
        if not scope:
            return self.log
        else:
            return self.log.getChild(".".join(scope))

    @classmethod
    @functools.cache
    def parse_parameters(cls) -> dict[str, Any]:
        parameters = cls.PARAM_DEFAULTS.copy()
        if cls.PARAM_FILE_PATH is None:
            logging.warning("No parameter file provided")
            return parameters
        path = Path(cls.PARAM_FILE_PATH)
        assert path.exists(), f"Parameter file does not exist: {path}"
        with path.open("r", encoding="utf-8") as fh:
            for key, value in json.load(fh).items():
                parameters[key.lower().strip()] = value
        return parameters

    @classmethod
    def get_parameter(cls, name: str, default: Any = None) -> Any:
        """
        Read back a parameter passed in from the outside world.

        :param name:    Name of the parameter to read
        :param default: Default value to return if parameter not defined
        :returns:       Value of the parameter or the default
        """
        return cls.parse_parameters().get(name.strip().lower(), default)

    async def ready(self) -> None:
        """Blocks until reset has completed"""
        await self.evt_ready.wait()
        self.evt_ready.clear()

    async def initialise(self) -> None:
        """Initialise the DUT's I/O"""
        self.rst.value = 1
        for comp in self._components.values():
            comp.io.initialise(IORole.opposite(comp.io.role))

    async def reset(self, init=True, wait_during=20, wait_after=1) -> None:
        """
        Reset the DUT.

        :param init:        Initialise the DUT's I/O
        :param wait_during: Clock cycles to hold reset active for (defaults to 20)
        :param wait_after:  Clock cycles to wait after lowering reset (defaults to 1)
        """
        # Drive reset high
        self.rst.value = 1
        # Initialise I/O
        if init:
            await self.initialise()
        # Wait before dropping reset
        if wait_during > 0:
            await ClockCycles(self.clk, wait_during)
        # Drop reset
        self.rst.value = 0
        # Wait for a bit
        if wait_after > 0:
            self.info(f"Waiting for {wait_after} cycles")
            await ClockCycles(self.clk, wait_after)

    def __getattr__(self, key: str) -> Any:
        """Pass through accesses to signals on the DUT.

        :param key: Name of the attribute
        :returns:   Resolved object
        """
        try:
            return getattr(super(), key)
        except Exception:
            return getattr(self.dut, key)

    def register(
        self,
        name: str,
        comp_or_coro: Component | Coroutine = None,
        scoreboard: bool = True,
        sb_queues: list[str] | tuple[str] | None = None,
        sb_filter: Callable | None = None,
        sb_timeout_ns: int | None = None,
        sb_polling_ns: int | None = None,
        sb_drain_policy: DrainPolicy | None = None,
        sb_match_window: int | None = None,
        **kwds,
    ) -> Component | Coroutine:
        """
        Register a driver, monitor, or coroutine with the testbench. Drivers and
        monitors must be provided a name and their random seeding will be setup
        from the testbench's random instance. Monitors will be registered with
        the scoreboard unless explicitly requested. Coroutines must also be named
        and are required to complete before the test will shutdown.

        :param name:            Name of the component or coroutine
        :param comp_or_coro:    Component instance or coroutine
        :param scoreboard:      Only applies to monitors, controls whether it is
                                registered with the scoreboard
        :param sb_queues:       A list of named queues used when a funnel type
                                scoreboard channel is required
        :param sb_filter:       A function that can filter or modify items
                                recorded by the monitor before they are passed
                                to the scoreboard
        :param sb_timeout_ns:   Optional timeout to allow for a object sat at the
                                front of the monitor queue to remain unmatched
                                (in nanoseconds, a value of None disables the
                                timeout mechanism)
        :param sb_polling_ns:   How frequently to poll to check for unmatched
                                items stuck in the monitor queue in nanoseconds
                                (defaults to 100 ns)
        :param sb_drain_policy: Specify how the handle should handle draining and
                                which queues should be considered when blocking
                                testcase shutdown
        :param sb_match_window: Where precise ordering of expected transactions
                                is not known, a positive integer matching window
                                can be used to match any of the next N transactions
                                in the reference queue (where N is set by
                                match_window)
        """
        # Allow either sb_ or scoreboard_ prefixes to arguments (back compatibility)
        sb_args = {
            "queues": sb_queues,
            "filter": sb_filter,
            "timeout_ns": sb_timeout_ns,
            "polling_ns": sb_polling_ns,
            "drain_policy": sb_drain_policy,
            "match_window": sb_match_window,
        }

        for key, value in sb_args.items():
            if (full_key := f"scoreboard_{key}") in kwds:
                assert value is None, f"Both sb_{key} and {full_key} specified"
                sb_args[key] = kwds.pop(full_key)

        assert not kwds, "Unexpected arguments: " + ", ".join(map(str, kwds.keys()))

        # Setup default argument values
        if sb_args["polling_ns"] is None:
            sb_args["polling_ns"] = 100
        if sb_args["drain_policy"] is None:
            sb_args["drain_policy"] = DrainPolicy.MON_AND_REF
        if sb_args["match_window"] is None:
            sb_args["match_window"] = 1

        assert isinstance(name, str), f"Name must be a string '{name}'"

        # Register a coroutine
        if asyncio.iscoroutine(comp_or_coro):
            assert name not in self._processes, f"Process known for '{name}'"
            task = cocotb.start_soon(comp_or_coro)
            self._processes[name] = task

        # Register a component (i.e. monitor or driver)
        elif isinstance(comp_or_coro, Component):
            assert name not in self._components, f"Component known for '{name}'"
            self._components[name] = comp_or_coro
            comp_or_coro.name = name
            setattr(self, name, comp_or_coro)
            comp_or_coro.seed(self.random)
            if scoreboard and isinstance(comp_or_coro, BaseMonitor):
                self.scoreboard.attach(
                    comp_or_coro,
                    filter_fn=sb_args["filter"],
                    queues=sb_args["queues"],
                    timeout_ns=sb_args["timeout_ns"],
                    polling_ns=sb_args["polling_ns"],
                    drain_policy=sb_args["drain_policy"],
                    match_window=sb_args["match_window"],
                )

        # Unknown type
        else:
            raise TypeError(f"Unsupported object: {comp_or_coro}")

        # Return the registered item to allow chaining
        return comp_or_coro

    def schedule(
        self,
        sequence: tuple[
            BaseSequence,
            Callable[[SimLog, random.Random, SeqArbiter, ModifiableObject, ModifiableObject], None],
        ],
        blocking: bool = True,
    ) -> Task:
        """
        Schedule a sequence to execute as part of a testcase.

        :param sequence: The sequence to schedule
        :param blocking: Whether the sequence must complete before the test is
                         allowed to finish
        :returns:        The scheduled task
        """
        seq_def, seq_inner = sequence
        task = cocotb.start_soon(
            seq_inner(
                self.fork_log("seq"),
                self.random,
                self._arbiter,
                self.clk,
                self.rst,
            )
        )
        if blocking:
            self._sequences.append((seq_def, task))
        return task

    def add_teardown(self, coro: Coroutine) -> None:
        """
        Register a coroutine to be executed after the shutdown loops have all
        completed, can be used to check final conditions.

        :param coro: Coroutine to register
        """
        assert asyncio.iscoroutine(coro), "Only coroutines may be added to teardown"
        self._teardown.append(coro)

    async def close_down(self, loops: int = 2, delay: int = 100) -> None:
        """
        Wait for drivers, monitors, and the scoreboard to drain to ensure that
        the test has completed.

        :param loops: Number of repetitions of the shutdown sequence to run
        :param delay: Number of clock cycles to wait between shutdown loops
        """
        # Filter drivers and monitors into separate lists
        drivers, monitors = [], []
        for comp in self._components.values():
            if not comp.blocking:
                continue
            [monitors, drivers][isinstance(comp, BaseDriver)].append(comp)
        # Check for consistent idleness
        for loop_idx in range(loops):
            # All drivers and monitors should be idle
            self._orch_log.info(f"Shutdown loop ({loop_idx+1}/{loops})")
            # Wait for sequences to complete
            self._orch_log.debug(f"Waiting for {len(self._sequences)} sequences to complete")
            for seq_def, seq_task in self._sequences:
                self._orch_log.debug(f"Waiting for sequence {seq_def.name}")
                await seq_task
            # Wait for minimum delay
            self._orch_log.debug("Waiting for minimum delay")
            await ClockCycles(self.clk, delay)
            # Wait for all drivers to return to idle
            for driver in drivers:
                self._orch_log.debug(f"Waiting for driver '{driver.name}' to go idle")
                await driver.idle()
            # Wait for all monitor queues to drain
            for monitor in monitors:
                self._orch_log.debug(f"Waiting for monitor '{monitor.name}' to go idle")
                await monitor.idle()
            # Wait for processes
            procs, self._processes = self._processes, {}
            for name, proc in procs.items():
                self._orch_log.debug(f"Waiting for process '{name}' to complete")
                await proc
            # Drain the scoreboard
            await self.scoreboard.drain()
        # Run teardown steps
        for teardown in self._teardown:
            await teardown

    @classmethod
    def testcase(
        cls,
        *args,
        reset=True,
        timeout=10000,
        shutdown_loops=2,
        shutdown_delay=100,
        reset_init=True,
        reset_wait_during=20,
        reset_wait_after=1,
        **kwargs,
    ) -> Callable:
        """
        Custom testcase declaration, wraps test with bench class

        :param reset:             Whether to reset the design
        :param timeout:           Maximum run time for a test (in clock cycles)
        :param shutdown_loops:    Number of loops of the shutdown sequence
        :param shutdown_delay:    Delay between loops of the shutdown sequence
        :param reset_init:        Initialise the DUT's I/O
        :param reset_wait_during: Clock cycles to hold reset active for
                                  (defaults to 20)
        :param reset_wait_after:  Clock cycles to wait after lowering reset
                                  (defaults to 1)
        """

        class _Testcase(cocotb.test):
            def __call__(self, dut, *args, **kwargs):
                async def _run_test():
                    # Clear components registered from previous runs
                    Component.COMPONENTS.clear()
                    # Create a testbench instance
                    try:
                        tb = cls(dut)
                    except Exception as e:
                        dut._log.error(f"Caught exception during {cls.__name__} constuction: {e}")
                        dut._log.error(traceback.format_exc())
                        raise e
                    # Log what's going on
                    tc_name = self._func.__name__
                    tb._orch_log.info(f"Preparing testcase {tc_name}")
                    # Check all components have been registered
                    missing = 0
                    for comp in Component.COMPONENTS:
                        if comp not in tb._components.values():
                            tb._orch_log.error(
                                f"{type(comp).__name__} '{comp.name}' has "
                                f"not been registered with the testbench"
                            )
                            missing += 1
                    assert missing == 0, "Some bench components have not been registered"
                    # If clock driving specified, start the clock
                    if tb.clk_drive:
                        cocotb.start_soon(Clock(tb.clk, tb.clk_period, units=tb.clk_units).start())
                    # If reset requested, run the sequence
                    if reset:
                        tb._orch_log.info("Resetting the DUT")
                        try:
                            await tb.reset(
                                init=reset_init,
                                wait_during=reset_wait_during,
                                wait_after=reset_wait_after,
                            )
                        except Exception as e:
                            tb._orch_log.error(f"Caught exception during reset: {e}")
                            tb._orch_log.error(traceback.format_exc())
                            raise e
                        tb._orch_log.info("DUT reset complete")

                    # Mark ready
                    tb.evt_ready.set()

                    # Wait for all components to be ready
                    for comp in tb._components.values():
                        await comp.ready()

                    # Create a forked log
                    log = tb.fork_log("test", tc_name)

                    # Are there any parameters for this test?
                    raw_tc_params = cls.get_parameter("testcases")
                    params = {}
                    for key, cast in cls.TEST_REQ_PARAMS[self._func]:
                        # First look for "<TESTCASE_NAME>.<PARAMETER_NAME>"
                        # but fall back to just "<PARAMETER_NAME>"
                        value = raw_tc_params.get(f"{tc_name}.{key}", raw_tc_params.get(key, None))
                        if value is None:
                            continue

                        # Cast string values to the appropriate type
                        if isinstance(value, str) and cast is not str:
                            if cast is bool:
                                value = strtobool(value)
                            else:
                                value = cast(value)

                        log.debug(f"Parameter {key}={value}")
                        params[key] = value

                    # Declare an intermediate function (this allows us to wrap
                    # with a optional timeout)
                    async def _inner():
                        await self._func(tb, log, *args, **kwargs, **params)
                        await tb.close_down(loops=shutdown_loops, delay=shutdown_delay)

                    # Run with a timeout if specified
                    postponed = None
                    if timeout is None:
                        await _inner()
                    else:
                        try:
                            await with_timeout(_inner(), timeout, "ns")
                        except SimTimeoutError as e:
                            postponed = e
                            tb._orch_log.error(f"Simulation timed out after {timeout} ns")
                            # List any busy drivers
                            for name, driver in tb._components.items():
                                if isinstance(driver, BaseDriver) and driver.queued > 0:
                                    tb._orch_log.info(
                                        f"Driver {name} has {driver.queued} "
                                        f"items remaining in its queue"
                                    )

                    # Report status of scoreboard channels
                    for _, channel in tb.scoreboard.channels.items():
                        channel.report()

                    # When using postmortem, catch errors before exit
                    if tb.get_parameter("postmortem", False) and (
                        (postponed is not None) or not tb.scoreboard.result
                    ):
                        tb._orch_log.warning("Entering postmortem")
                        breakpoint()

                    # If an exception has been postponed, re-raise it now
                    if postponed is not None:
                        raise postponed

                    # Check the result
                    assert tb.scoreboard.result, "Scoreboard reported test failure"

                return cocotb.decorators._RunningTest(_run_test(), self)

        def _do_decorate(func):
            # _testcase acts as a function which returns a decorator, hence the
            # double function call
            return _Testcase(*args, **kwargs)(func)

        return _do_decorate

    @classmethod
    def parameter(cls, name: str, cast: Callable[[str], Any]) -> Callable:
        """
        Decorator for defining a parameter of a testcase that can be overridden
        from a parameter file identified by the `TEST_PARAMS` environment
        variable.

        :param name: Name of the parameter
        :param cast: Function to cast a string param to the required type,
                     usually just the type constructor e.g. `int`, `float`.
                     As a special case `bool` will match intuitively based
                     on string content, e.g. `True, y, False, NO, 1, 0, off`.
        """

        def _inner(method: Callable) -> Callable:
            cls.TEST_REQ_PARAMS[method].append((name, cast))
            return method

        return _inner


# Start profiling when it is enabled in the parameters file
if outfile := BaseBench.get_parameter("profiling"):
    import atexit

    import yappi

    logging.warning("Profiling has been enabled")
    yappi.set_clock_type("wall")
    yappi.start()

    # Register a teardown method to stop profiling when Python exits
    def _end_profile():
        yappi.stop()
        logging.info("Profiling summary:")
        logging.info(yappi.get_func_stats().print_all())
        logging.info(f"Profile data written to {outfile}")
        yappi.get_func_stats().save(outfile, type="pstat")

    atexit.register(_end_profile)



# Copyright 2023, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from enum import Enum, auto
from random import Random
from typing import Any

import cocotb
from cocotb.handle import ModifiableObject
from cocotb.queue import Queue
from cocotb.triggers import Event, RisingEdge
from cocotb.utils import get_sim_time

from .component import Component
from .io import BaseIO
from .transaction import BaseTransaction


class DriverEvent(Enum):
    ENQUEUE = auto()
    """Emitted when a transaction is enqueued to a driver"""
    PRE_DRIVE = auto()
    """Emitted just prior to a queued transaction being driven into the DUT"""
    POST_DRIVE = auto()
    """Emitted just after a queued transaction has been driven into the DUT"""


@dataclasses.dataclass()
class DriverStatistics:
    dequeued: int = 0


class BaseDriver(Component):
    """
    Component for driving transactions onto an interface matching the
    implementation's signalling protocol.

    :param tb:      Handle to the testbench
    :param io:      Handle to the BaseIO interface
    :param clk:     Clock signal to use when driving/sampling the interface
    :param rst:     Reset signal to use when driving/sampling the interface
    :param random:  Random number generator to use (optional)
    :param name:    Unique name for this component instance (optional)
    """

    def __init__(
        self,
        tb: Any,
        io: BaseIO,
        clk: ModifiableObject,
        rst: ModifiableObject,
        random: Random | None = None,
        name: str | None = None,
        blocking: bool = True,
    ) -> None:
        super().__init__(tb, io, clk, rst, random, name, blocking)
        self.stats = DriverStatistics()
        self._queue: Queue[BaseTransaction] = Queue()
        cocotb.start_soon(self._driver_loop())

    @property
    def busy(self) -> bool:
        """Busy when either locked or the queue has outstanding entries"""
        return not self._queue.empty() and super().busy

    @property
    def queued(self) -> int:
        """Return how many entries are queued up"""
        return self._queue.qsize()

    def enqueue(
        self, transaction: BaseTransaction, wait_for: DriverEvent | None = None
    ) -> Event | None:
        """
        Queue up a transaction to be driven onto the interface

        :param transaction: Transaction to queue, must inherit from BaseTransaction
        :param wait_for:    When defined, this will return an event that can be
                            monitored for a given transaction event occurring
        """
        # Sanity check
        if not isinstance(transaction, BaseTransaction):
            raise TypeError(
                f"Transaction objects should inherit from BaseTransaction unlike {transaction}"
            )
        # Does this transaction need an event?
        if wait_for is not None:
            transaction._f_event = wait_for
            transaction._c_event = Event()
        # Queue up the transaction with no delay
        self._queue.put_nowait(transaction)
        # Notify any enqueue subscribers
        self.publish(DriverEvent.ENQUEUE, transaction)
        if transaction._f_event is DriverEvent.ENQUEUE:
            transaction._c_event.set()
        # Return the cocotb Event (if it was set)
        return transaction._c_event

    async def _driver_loop(self) -> None:
        """Main loop for driving transactions onto the interface"""
        await self.tb.ready()
        await RisingEdge(self.clk)
        self._ready.set()
        while True:
            # Pickup next event to drive
            obj = await self._queue.get()
            # Wait until reset is deasserted
            while self.rst.value == 1:
                await RisingEdge(self.clk)
            # Lock out the driver (prevents shutdown mid-stimulus)
            await self.lock()
            # Set the timestamp where the transaction was about to be driven
            obj.timestamp = get_sim_time(units="ns")
            # Notify any pre-drive subscribers
            self.publish(DriverEvent.PRE_DRIVE, obj)
            if obj._f_event is DriverEvent.PRE_DRIVE:
                obj._c_event.set()
            # Drive the transaction
            await self.drive(obj)
            self.stats.dequeued += 1
            # Notify any post-drive subscribers
            self.publish(DriverEvent.POST_DRIVE, obj)
            if obj._f_event is DriverEvent.POST_DRIVE:
                obj._c_event.set()
            # Release the lock
            self.release()

    async def drive(self, obj: BaseTransaction) -> None:
        """
        Placeholder driver, this should be overridden by a child class to match
        the signalling protocol of the interface's implementation.

        :param obj: The transaction to drive onto the interface
        """
        del obj
        raise NotImplementedError("drive is not implemented on BaseDriver")



# Copyright 2023, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from random import Random
from typing import Any, ClassVar

from cocotb.handle import ModifiableObject
from cocotb.triggers import RisingEdge

from .event import EventEmitter
from .io import BaseIO


class Component(EventEmitter):
    """
    Base component type for working with BaseIO interfaces, can be extended to
    form drivers, monitors, and other signalling protocol aware components.

    :param tb:       Handle to the testbench
    :param io:       Handle to the BaseIO interface
    :param clk:      Clock signal to use when driving/sampling the interface
    :param rst:      Reset signal to use when driving/sampling the interface
    :param random:   Random number generator to use (optional)
    :param name:     Unique name for this component instance (optional)
    :param blocking: Whether this component should block shutdown (default: True)
    """

    # Tracks all component instances
    COMPONENTS: ClassVar[list["Component"]] = []

    def __init__(
        self,
        tb: Any,
        io: BaseIO,
        clk: ModifiableObject,
        rst: ModifiableObject,
        random: Random | None = None,
        name: str | None = None,
        blocking: bool = True,
    ) -> None:
        # To avoid an import loop
        from .bench import BaseBench

        assert isinstance(tb, BaseBench), "'tb' should inherit from BaseBench"
        assert isinstance(io, BaseIO), "'io' should inherit from BaseIO"
        # Setup emitter behaviours
        super().__init__()
        # Setup component variables
        self.tb = tb
        self.io = io
        self.clk = clk
        self.rst = rst
        self.random = Random(random.random() if random else 0)
        self.name = name or type(self).__name__
        self.blocking = blocking
        self.log = self.tb.fork_log(type(self).__name__.lower(), self.name)
        self._lock = asyncio.Lock()
        Component.COMPONENTS.append(self)

    async def ready(self) -> None:
        await self._ready.wait()

    def seed(self, random: Random) -> None:
        """
        Set up the random seed (used by testbench when registering a component)

        :param random:  The random instance to seed from
        """
        self.random = Random(random.random())

    async def lock(self) -> None:
        """Lock the component's internal lock"""
        await self._lock.acquire()

    def release(self) -> None:
        """Release the component's internal lock"""
        self._lock.release()

    @property
    def busy(self) -> bool:
        """Determine if the component is currently busy"""
        return self._lock.locked()

    async def idle(self) -> None:
        """Blocks until the component is no longer busy"""
        while True:
            await RisingEdge(self.clk)
            if not self.busy:
                break



# Copyright 2023, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import itertools
import logging
from collections import defaultdict
from collections.abc import Callable, Iterable
from enum import Enum, auto
from random import Random
from typing import Any, ClassVar, Generic, Self, TypeVar

import cocotb
from cocotb.handle import ModifiableObject
from cocotb.log import SimLog
from cocotb.triggers import Event, First, Lock

from .component import Component
from .driver import BaseDriver
from .event import EventEmitter


class SeqLock:
    """
    Wraps around cocotb's Lock primitive to also track which sequence currently
    holds the lock.

    :param name: Name of the lock
    """

    # Locks for components
    _COMPONENT_LOCKS: ClassVar[dict[Component, Self]] = {}
    # Named locks
    _NAMED_LOCKS: ClassVar[dict[str, Self]] = {}

    def __init__(self, name: str) -> None:
        self._name = name
        self._lock = Lock(name)
        self._locked_by: BaseSequence | None = None

    @classmethod
    def get_component_lock(cls, comp: Component) -> Self:
        """
        Retrieve the shared lock for a specific component.

        :param comp: Reference to the component
        :returns:    The shared lock instance
        """
        if comp in cls._COMPONENT_LOCKS:
            return cls._COMPONENT_LOCKS[comp]
        else:
            cls._COMPONENT_LOCKS[comp] = (lock := SeqLock(comp.name))
            return lock

    @classmethod
    def get_named_lock(cls, name: str) -> Self:
        """
        Retrieve a shared named lock (distinct from all component locks).

        :param name: Name of the shared lock
        :returns:    The shared lock instance
        """
        if name in cls._NAMED_LOCKS:
            return cls._NAMED_LOCKS[name]
        else:
            cls._NAMED_LOCKS[name] = (lock := SeqLock(name))
            return lock

    @classmethod
    def get_all_component_locks(cls) -> Iterable[tuple[Component, Self]]:
        """Return a list of all known component locks"""
        yield from cls._COMPONENT_LOCKS.items()

    @classmethod
    def get_all_named_locks(cls) -> Iterable[tuple[str, Self]]:
        """Return a list of all known named locks"""
        yield from cls._NAMED_LOCKS.items()

    @classmethod
    def get_all_locks(cls) -> Iterable[Self]:
        """Return a list of all known locks"""
        yield from cls._COMPONENT_LOCKS.values()
        yield from cls._NAMED_LOCKS.values()

    @classmethod
    def count_all_locks(cls) -> Iterable[Self]:
        """Return the total number of locks"""
        return len(cls._COMPONENT_LOCKS) + len(cls._NAMED_LOCKS)

    @property
    def locked(self) -> bool:
        """Check if the lock is currently taken"""
        # NOTE: Prior to 1.8.0 of cocotb 'locked' was an attribute, now it is
        #       a function that must be called
        return self._lock.locked() if callable(self._lock.locked) else self._lock.locked

    async def acquire(self, context: "SeqContext") -> None:
        """
        Attempt to acquire a lock, waiting until it becomes available.

        :param context: Reference to the sequence context claiming the lock
        """
        assert isinstance(context, SeqContext)
        if self._lock.locked and self._locked_by is context:
            return
        else:
            await self._lock.acquire()
            self._locked_by = context

    def release(self, context: "SeqContext") -> None:
        """
        Release a held lock only if the context matches the current lock holder.

        :param context: Reference to the sequence context that previously claimed
                        the lock
        """
        assert isinstance(context, SeqContext)
        if self._lock.locked:
            assert self._locked_by is context, f"Illegal attempt to unlock {self._name}"
            self._locked_by = None
            self._lock.release()
            # Raise unlock event
            SeqContext.SEQ_SHARED_EVENT.publish(SeqContextEvent.UNLOCKED, None)


class SeqRandomVariable:
    """
    Captures the definition of a random variable used in a sequence along with
    how it should be randomised.

    :param name:      Name of the variable
    :param bit_width: Randomise over a given number of bits
    :param range:     Randomise over a given range (int or float)
    :param choices:   Make a random selection from a list of choices
    """

    SUFFIXES: ClassVar[tuple[str]] = ("bit_width", "range", "choices")

    def __init__(
        self,
        name: str,
        bit_width: int | None = None,
        range: tuple[int | float, int | float] | None = None,  # noqa: A002
        choices: tuple[Any] | None = None,
    ) -> None:
        # Take references
        self.name = name.strip().replace(" ", "_")
        self.bit_width = bit_width
        self.range = range
        self.choices = choices
        # Sanity checks
        assert len(name) > 0, "Variable name cannot be an empty string"
        assert not name.endswith("_bitwidth"), "Variable name must not end with '_bitwidth'"
        assert not name.endswith("_range"), "Variable name must not end with '_range'"
        assert not name.endswith("_choices"), "Variable name must not end with '_choices'"
        assert (
            len([x for x in (bit_width, range, choices) if x is not None]) <= 1
        ), "Only one of bit width, range, or choices may be specified"
        if bit_width is not None:
            assert (
                isinstance(bit_width, int) and bit_width > 0
            ), "Bit width must be an integer value greater than zero"
        elif range is not None:
            assert (
                isinstance(range, tuple) and len(range) == 2
            ), "Range must be a tuple of two values"
            assert all(
                isinstance(x, int | float) for x in range
            ), "All entries of the range must be integer or float"
        elif choices is not None:
            assert (
                isinstance(choices, tuple) and len(choices) > 0
            ), "Choices must be a tuple of at least one value"

    @property
    def varnames(self) -> tuple[str]:
        return [self.name] + [f"{self.name}_{s}" for s in self.SUFFIXES]

    def randomise(
        self,
        random: Random,
        bit_width: int | None = None,
        range: tuple[int | float, int | float] | None = None,  # noqa: A002
        choices: tuple[Any] | None = None,
    ) -> Any:
        # If any overrides are provided, evaluate them first
        if bit_width is not None:
            return random.getrandbits(bit_width)
        elif range is not None:
            if all(isinstance(x, int) for x in range):
                return random.randrange(*range)
            else:
                return random.uniform(*range)
        elif choices is not None:
            return random.choice(choices)
        # Otherwise evaluate default behaviour
        if self.bit_width is not None:
            return random.getrandbits(self.bit_width)
        elif self.range is not None:
            if all(isinstance(x, int) for x in self.range):
                return random.randrange(*self.range)
            else:
                return random.uniform(*self.range)
        elif self.choices is not None:
            return random.choices(self.choices)
        # Should not get here!
        raise Exception("Failed to select a randomisation method")


C = TypeVar("C")


class SeqProxy(EventEmitter, Generic[C]):
    """
    Wraps around a component to provide locking and masking functionality, this
    is achieved by mocking functionality of a component including intercepting
    and filtering events.

    :param context:   The sequence context associated to this proxy
    :param component: The component to proxy
    :param lock:      The shared lock for the component
    """

    def __init__(self, context: "SeqContext", component: C, lock: SeqLock) -> None:
        super().__init__()
        assert isinstance(component, Component)
        self._context = context
        self._component = component
        self._lock = lock
        # Subscribe to events from the component
        self._component.subscribe_all(self._event_callback)

    @property
    def _holds_lock(self) -> bool:
        return self._lock._lock.locked and self._lock._locked_by is self._context

    def _event_callback(self, comp: Component, event: Enum, obj: Any) -> None:
        """
        Intercepts all events coming from the component and drops them if the
        lock is held by another sequence.

        :param comp:  Component emitting the event
        :param event: The event being emitted
        :param obj:   The event object
        """
        if not self._lock._lock.locked or self._holds_lock:
            self.publish(event, obj)

    def enqueue(self, *args: Any, **kwds: Any) -> Event | None:
        """
        Forward an enqueue request through from the proxy to the wrapped driver.

        :param *args: Arguments to forward
        :param *kwds: Keyword arguments to forward
        """
        if isinstance(self._component, BaseDriver):
            if not self._holds_lock:
                raise Exception(
                    f"Attempting to enqueue into {type(self._component).__name__} "
                    f"without first acquiring the lock, instead the lock is held "
                    f"by {self._lock._locked_by}"
                )
            return self._component.enqueue(*args, **kwds)
        else:
            raise Exception(f"Cannot enqueue to '{type(self._component).__name__}'")

    @property
    def clk(self) -> ModifiableObject:
        return self._component.clk

    @property
    def rst(self) -> ModifiableObject:
        return self._component.rst

    @property
    def io(self) -> ModifiableObject:
        return self._component.io

    def idle(self) -> None:
        """Forward idle through to the wrapped component"""
        return self._component.idle()


class SeqContextEvent(Enum):
    UNLOCKED = auto()


class SeqArbiter:
    """
    Arbitrates being queuing sequences to determine which sequences can start
    based on the locks they are requesting.
    """

    def __init__(self, log: SimLog, random: Random):
        self._log = log
        self._debug = log.getEffectiveLevel() <= logging.DEBUG
        self._random = Random(random.random())
        self._queue = []
        self._evt_queue = Event()
        cocotb.start_soon(self._manage())

    async def queue_for(self, context: "SeqContext", locks: list[SeqLock]) -> None:
        """
        Queue against the arbiter for a collection of locks. The arbiter will
        schedule the sequence only once all locks can be atomically acquired.

        :param context: The sequence context queueing
        :param locks:   The list of locks required
        """
        # Queue up the context, locks it requests, and the stall event
        self._queue.append((context, locks, evt := Event()))
        # Mark a new entry as having been pushed
        self._evt_queue.set()
        # Wait for the event
        await evt.wait()

    async def _manage(self) -> None:
        """Executes in a loop to schedule sequences"""
        for idx in itertools.count():
            # Wait until something is queued up
            await self._evt_queue.wait()
            # While stuff is queued, attempt to schedule it
            while self._queue:
                # Keep track of which locks are available
                available = {x for x in SeqLock.get_all_locks() if not x.locked}
                # If no locks are available, wait for the next release
                if not available:
                    await SeqContext.SEQ_SHARED_EVENT.wait_for(SeqContextEvent.UNLOCKED)
                    continue
                # Randomise the order to process the queue
                order = list(range(len(self._queue)))
                self._random.shuffle(order)
                # Log the scheduling problem
                pre_queue, pre_free = len(self._queue), available.copy()
                # Schedule as many sequences as possible
                scheduled = []
                for idx in order:
                    # Pickup the entry
                    ctx, locks, evt = self._queue[idx]
                    # If any locks are unavailable, keep searching
                    if set(locks).difference(available):
                        continue
                    # Claim the locks on behalf of the sequence
                    for lock in locks:
                        await lock.acquire(ctx)
                        available.remove(lock)
                    # Trigger the sequence release event
                    evt.set()
                    # Remember which sequences have been scheduled
                    scheduled.append(idx)
                    # If no more locks are available, break out early
                    if not available:
                        break
                # Log what was scheduled when in debug mode
                if self._debug and (post_sched := len(scheduled)) > 0:
                    post_claim, post_diff = locks[:], pre_free.difference(locks)
                    msg = (
                        f"Scheduled {post_sched} "
                        f"({', '.join(self._queue[x][0].id for x in scheduled)}) "
                        f"out of {pre_queue} sequences, claiming {len(post_claim)} "
                        f"({', '.join(x._name for x in post_claim)}) locks"
                    )
                    if len(post_diff) > 0:
                        self._log.debug(
                            f"{msg} leaving {len(post_diff)} "
                            f"({', '.join(x._name for x in post_diff)})"
                        )
                    else:
                        self._log.debug(msg)
                # Prune the scheduled sequences
                self._queue = [x for i, x in enumerate(self._queue) if i not in scheduled]
                # If scheduling was unsuccessful then either wait for locks to
                # be released or new sequences to be scheduled
                # NOTE: As the scheduling loop contains an `await` it is
                #       possible for new sequences to be queued up after the
                #       initial "order" is determined. This is why the code below
                #       only waits if no sequences were successfully scheduled.
                if not scheduled:
                    # Clear trigger event so that the next queue_for raises it
                    self._evt_queue.clear()
                    await First(
                        SeqContext.SEQ_SHARED_EVENT._get_wait_event(
                            SeqContextEvent.UNLOCKED
                        ).wait(),
                        self._evt_queue.wait(),
                    )
            # Clear the trigger event so that the next queue_for call retriggers
            # the scheduling routine
            self._evt_queue.clear()


class SeqContext:
    """
    A context specific to a given sequence invocation that provides logging,
    random value generation, and lock management.

    :param sequence: The sequence being invoked
    :param log:      The root sequencing log (pre-fork)
    :param random:   The root random instance (pre-fork)
    """

    SEQ_CTX_ID: ClassVar[dict[str, itertools.count]] = defaultdict(itertools.count)
    SEQ_SHARED_LOCK: ClassVar[Lock] = Lock()
    SEQ_SHARED_EVENT: ClassVar[EventEmitter] = EventEmitter()

    def __init__(
        self,
        sequence: "BaseSequence",
        log: SimLog,
        random: Random,
        arbiter: SeqArbiter,
        clk: ModifiableObject,
        rst: ModifiableObject,
    ) -> None:
        self._sequence = sequence
        self._arbiter = arbiter
        # Allocate an ID unique to the sequence's name
        self._ctx_id = next(type(self).SEQ_CTX_ID[self._sequence.name])
        # Fork the log to ensure a distinct context
        self.log = log.getChild(self.id)
        # Fork the root random to ensure sequence run-to-run consistency
        self.random = Random(random.random())
        # Reference clock and reset
        self.clk = clk
        self.rst = rst
        # Lock re-entrancy flag
        self._locks_active = False

    @property
    def id(self) -> str:
        return f"{self._sequence.name}[{self._ctx_id}]"

    @contextlib.asynccontextmanager
    async def lock(self, *lockables: SeqLock | SeqProxy):
        """
        Atomically acquire one or more named or component locks (i.e. locks will
        only be claimed if all locks are available, otherwise it will wait until
        such a condition can be met).

        :param *lockables: References to named locks or sequence proxies (which
                           will be resolved to the equivalent component locks)
        """
        # Mark that locking is active
        assert not self._locks_active, "You must release all locks before re-acquisition"
        self._locks_active = True
        # Figure out all the locks that need to be acquired
        need: list[SeqLock] = []
        for lock in lockables:
            if isinstance(lock, SeqProxy):
                need.append(SeqLock.get_component_lock(lock._component))
            else:
                need.append(lock)
        # Queue against the arbiter for a slot to execute
        await self._arbiter.queue_for(self, need)
        # Yield to allow the sequence to execute
        yield
        # Release any remaining locks
        for lock in need:
            if lock._locked_by is self:
                lock.release(self)
        # Clear locking active flag
        self._locks_active = False

    def release(self, *lockables: SeqLock | SeqProxy):
        """
        Release one or more named or component locks, allowing other sequences
        to move forward. Note that attempting to release a lock that the sequence
        doesn't hold will raise an exception.

        :param *lockables: References to named locks or sequence proxies (which
                           will be resolved to the equivalent component locks)
        """
        for lock in lockables:
            if isinstance(lock, SeqProxy):
                SeqLock.get_component_lock(lock._component).release(self)
            else:
                lock.release(self)


class BaseSequence:
    """
    Wraps around a sequencing function and services lock requests and other
    integration with the testbench.

    :param fn: The sequencing function being wrapped
    """

    REGISTRY: ClassVar[dict[Callable, "BaseSequence"]] = {}
    LOCKS: ClassVar[dict[str, SeqLock]] = defaultdict(SeqLock)

    def __init__(self, fn: Callable, auto_lock: bool = False) -> None:
        # Ensure that this wrapper is unique for the sequencing function
        assert id(fn) not in type(self).REGISTRY, "Sequencing function registered twice"
        type(self).REGISTRY[id(fn)] = self
        # Capture variables
        self.auto_lock = auto_lock
        self._fn = fn
        self._requires: dict[str, Any] = {}
        self._locks: list[str] = []
        self._randargs: dict[str, SeqRandomVariable] = {}

    def __repr__(self) -> str:
        return f'<BaseSequence name="{self.name}">'

    @property
    def name(self) -> str:
        return self._fn.__name__

    @classmethod
    def register(cls, fn: Callable | Self) -> Self:
        """
        Uniquely wrap a sequencing function inside a BaseSequence, returning the
        shared instance on future invocations.

        :param fn: The sequencing function
        :returns:  The wrapping BaseSequence instance
        """
        if isinstance(fn, BaseSequence):
            return fn
        elif id(fn) in cls.REGISTRY:
            return cls.REGISTRY[id(fn)]
        else:
            return BaseSequence(fn)

    def add_requirement(self, req_name: str, req_type: Any) -> Self:
        """
        Add a requirement on a driver/monitor that must be provided by the
        testbench for the sequence to execute.

        :param req_name: Name of the driver/monitor
        :param req_type: Type of the driver/monitor
        :returns:        Self to allow for simple chaining
        """
        req_name = req_name.strip().lower().replace(" ", "_")
        assert len(req_name) > 0, "Requirement name cannot be an empty string"
        assert req_name not in self._requires, f"Requirement already placed on {req_name}"
        assert req_name not in self._locks, f"Requirement {req_name} clashes with a lock"
        self._requires[req_name] = req_type
        return self

    def add_lock(self, lock_name: str) -> Self:
        """
        Define an arbitrary lock to support simple cross-sequence co-ordination
        over resources internal to the design (that are not well captured by a
        driver or monitor requirement).

        :param lock_name: Name of the lock
        :returns:         Self to allow for simple chaining
        """
        lock_name = lock_name.strip().lower().replace(" ", "_")
        assert len(lock_name) > 0, "Lock name cannot be an empty string"
        assert lock_name not in self._locks, f"Lock '{lock_name}' has already been defined"
        assert lock_name not in self._requires, f"Lock {lock_name} clashes with a requirement"
        self._locks.append(lock_name)
        return self

    def add_randarg(
        self,
        name: str,
        bit_width: int | None = None,
        range: tuple[int | float, int | float] | None = None,  # noqa: A002
        choices: tuple[Any] | None = None,
    ) -> Self:
        """
        Define an argument that can be randomised in a number of different ways.
        Only one method of randomisation may be specified.

        :param name:      Name of the argument
        :param bit_width: Randomise over a given number of bits
        :param range:     Randomise over a given range (int or float)
        :param choices:   Make a random selection from a list of choices
        :returns:         Self to allow for simple chaining
        """
        rv = SeqRandomVariable(name, bit_width, range, choices)
        self._randargs[rv.name] = rv
        return self

    def __call__(self, **kwds):
        """
        Call the underlying sequence with any parameters that it requires, and
        launch prepare it to be launched within a managed context.

        :param **kwds: Any keyword arguments
        :returns:      The wrapped coroutine
        """

        # Create a wrapper to allow log and random to be inserted by the bench
        async def _inner(
            log: SimLog,
            random: Random,
            arbiter: SeqArbiter,
            clk: ModifiableObject,
            rst: ModifiableObject,
        ):
            # Create a context
            ctx = SeqContext(self, log, random, arbiter, clk, rst)
            # Check that provided components match expectation
            comps = {}
            for name, ctype in self._requires.items():
                # Check for a missing keyword argument
                if name not in kwds:
                    raise Exception(f"No component provided for '{name}'")
                # Check if the expected component type matches
                # NOTE: Delete from kwds to avoid clash during expansions
                match = kwds[name]
                del kwds[name]
                if not isinstance(match, ctype):
                    raise Exception(f"Component '{name}' is not of type {ctype.__name__}")
                # Ensure a component lock exists
                comp_lock = SeqLock.get_component_lock(match)
                # Pickup the component and wrap it in a proxy
                comps[name] = SeqProxy(ctx, match, comp_lock)
            # Generate named locks
            locks = {}
            for lock in self._locks:
                locks[lock] = SeqLock.get_named_lock(lock)
            # Identify random arguments and overrides
            arguments = {}
            for name, rv in self._randargs.items():
                # Identify any matching overrides
                overrides = [(k, v) for k, v in kwds.items() if k in rv.varnames]
                assert len(overrides) in (0, 1), (
                    f"Expecting a maximum of one override for {name}, instead "
                    f"found: {', '.join(x[0] for x in overrides)}"
                )
                # If not overridden, immediately randomise
                if not overrides:
                    arguments[name] = rv.randomise(ctx.random)
                    continue
                # Determine override mode
                o_name, o_val = overrides[0]
                for sfx in SeqRandomVariable.SUFFIXES:
                    if o_name.endswith("_" + sfx):
                        arguments[name] = rv.randomise(ctx.random, **{sfx: o_val})
                        break
                # If no override mode was matched, this is a fixed value
                else:
                    arguments[name] = o_val
                # Delete from kwds
                del kwds[o_name]
            # Fill in all remaining values
            for name, value in kwds.items():
                arguments[name] = value
            # Log what's about to launch
            ctx.log.debug(f"Launching {ctx.id} with arguments: {arguments}")
            # If auto-locking requested, wrap with a lock context
            if self.auto_lock:
                async with ctx.lock(*comps.values(), *locks.values()):
                    await self._fn(ctx, **comps, **locks, **arguments)
            # Otherwise just launch the sequence directly
            else:
                await self._fn(ctx, **comps, **locks, **arguments)

        # Return wrapped coroutine
        return self, _inner


def sequence(auto_lock: bool = False) -> BaseSequence:
    """
    Decorator used to wrap a sequencing function, for now there are no arguments
    and the argument pattern is just a placeholder for future extension.

    :param auto_lock: When enabled locks will be claimed automatically on all
                      requirements as the sequence starts, otherwise the sequence
                      will be responsible for claiming and releasing locks
    :returns:         Wrapped sequence
    """

    def _inner(fn: Callable) -> Callable:
        seq = BaseSequence.register(fn)
        seq.auto_lock = auto_lock
        return seq

    return _inner


def requires(req_name: str, req_type: Any | None = None) -> BaseSequence:
    """
    Decorator used to add a requirement on a driver/monitor or an arbitrarily
    named lock to a sequencing function.

    :param req_name: Name of the lock, driver, or monitor required
    :param req_type: Type of the driver/monitor required
    :returns:        Wrapped sequence
    """

    def _inner(fn: Callable) -> Callable:
        if req_type is None:
            return BaseSequence.register(fn).add_lock(req_name)
        else:
            return BaseSequence.register(fn).add_requirement(req_name, req_type)

    return _inner


def randarg(
    name: str,
    bit_width: int | None = None,
    range: tuple[int | float, int | float] | None = None,  # noqa: A002
    choices: tuple[Any] | None = None,
) -> BaseSequence:
    """
    Decorator used to add a randomised argument to a sequence definition, that
    can be randomised in a number of different ways. Only one method of
    randomisation may be specified.

    :param name:      Name of the argument
    :param bit_width: Randomise over a given number of bits
    :param range:     Randomise over a given range (int or float)
    :param choices:   Make a random selection from a list of choices
    :returns:         Wrapped sequence
    """

    def _inner(fn: Callable) -> Callable:
        return BaseSequence.register(fn).add_randarg(
            name=name, bit_width=bit_width, range=range, choices=choices
        )

    return _inner



# Copyright 2023, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from collections.abc import Callable
from enum import Enum, auto
from random import Random
from typing import Any

import cocotb
from cocotb.handle import ModifiableObject
from cocotb.triggers import RisingEdge

from .component import Component
from .io import BaseIO
from .transaction import BaseTransaction


class MonitorEvent(Enum):
    CAPTURE = auto()


@dataclasses.dataclass()
class MonitorStatistics:
    captured: int = 0


class BaseMonitor(Component):
    """
    Component for sampling transactions from an interface matching the
    implementation's signalling protocol.

    :param tb:      Handle to the testbench
    :param io:      Handle to the BaseIO interface
    :param clk:     Clock signal to use when driving/sampling the interface
    :param rst:     Reset signal to use when driving/sampling the interface
    :param random:  Random number generator to use (optional)
    :param name:    Unique name for this component instance (optional)
    """

    def __init__(
        self,
        tb: Any,
        io: BaseIO,
        clk: ModifiableObject,
        rst: ModifiableObject,
        random: Random | None = None,
        name: str | None = None,
        blocking: bool = True,
    ) -> None:
        super().__init__(tb, io, clk, rst, random, name, blocking)
        self.stats = MonitorStatistics()
        cocotb.start_soon(self._monitor_loop())

    async def _monitor_loop(self) -> None:
        """Main loop for monitoring transactions on the interface"""
        await self.tb.ready()
        await RisingEdge(self.clk)
        self._ready.set()

        def _capture(obj: BaseTransaction):
            self.stats.captured += 1
            self.publish(MonitorEvent.CAPTURE, obj)

        while True:
            await self.monitor(_capture)

    async def monitor(self, capture: Callable) -> None:
        """
        Placeholder monitor, this should be overridden by a child class to match
        the signalling protocol of the interface's implementation.

        :param capture: Function to call whenever a transaction is captured
        """
        del capture
        raise NotImplementedError("monitor is not implemented on BaseMonitor")



# Copyright 2023, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from enum import IntEnum
from typing import Any

from cocotb.handle import HierarchyObject, NonHierarchyObject
from cocotb.log import SimLog


class IORole(IntEnum):
    """Role that a particular bus is performing (determines signal suffix)"""

    INITIATOR = 0
    RESPONDER = 1

    @staticmethod
    def opposite(value: "IntEnum") -> "IntEnum":
        return {IORole.INITIATOR: IORole.RESPONDER, IORole.RESPONDER: IORole.INITIATOR}[value]


class SignalWrapper:
    """
    Some simulators supported by cocotb give visibility into the fields of a
    struct (e.g. Cadence Xcelium), while others only expose it as a packed
    bitvector. Forastero is built to assume the signal is a packed bitvector,
    and this wrapper class normalises the behaviour.

    :param hier: The signal hierarchy object
    """

    def __init__(self, hier: HierarchyObject | NonHierarchyObject) -> None:
        self._hier = hier

        # Recursively discover all components (in case of a nested struct)
        def _expand(level):
            if isinstance(level, HierarchyObject):
                for comp in level:
                    yield from _expand(comp)
            else:
                yield level

        all_components = list(_expand(self._hier))
        # Figure out how the bit fields pack into the bit vector, precalculating
        # the MSB, LSB, and mask to save compute later
        self._packing = []
        self._width = 0
        for comp in all_components[::-1]:
            if comp._range is None:
                c_msb, c_lsb = len(comp) - 1, 0
            else:
                c_msb, c_lsb = comp._range
            rel_msb, rel_lsb = self._width + c_msb, self._width + c_lsb
            width = len(comp)
            self._packing.append(((rel_lsb, rel_msb, (1 << width) - 1), comp))
            self._width += width

    @property
    def value(self) -> int:
        value = 0
        for (lsb, _, _), comp in self._packing:
            value |= int(comp.value) << lsb
        return value

    @value.setter
    def value(self, value: int) -> None:
        for (lsb, _, mask), comp in self._packing:
            comp.value = (value >> lsb) & mask

    @property
    def _range(self) -> tuple[int, int]:
        return 0, self._width - 1

    def __len__(self) -> int:
        return self._width


def io_prefix_style(bus: str | None, component: str, role_bus: IORole, role_comp: IORole) -> str:
    """
    Style signal names as i/o_(<BUS>_)<COMPONENT> for example i_dma_awaddr and
    o_dma_awready.

    :param bus:       Name of the bus instance or None if not required
    :param component: Name of component signal withim the bus
    :param role_bus:  Interface role of the entire bus instance
    :param role_comp: Interface role of the component within the bus
    :returns:         Complete string name
    """
    mapping = {
        (IORole.INITIATOR, IORole.INITIATOR): "o",
        (IORole.INITIATOR, IORole.RESPONDER): "i",
        (IORole.RESPONDER, IORole.INITIATOR): "i",
        (IORole.RESPONDER, IORole.RESPONDER): "o",
    }
    full_name = f"{mapping[role_bus, role_comp]}"
    if bus is not None:
        full_name += f"_{bus}"
    return f"{full_name}_{component}"


def io_suffix_style(bus: str | None, component: str, role_bus: IORole, role_comp: IORole) -> str:
    """
    Style signal names as (<BUS>_)<COMPONENT>_i/o for example dma_awaddr_i and
    dma_awready_o.

    :param bus:       Name of the bus instance or None if not required
    :param component: Name of component signal withim the bus
    :param role_bus:  Interface role of the entire bus instance
    :param role_comp: Interface role of the component within the bus
    :returns:         Complete string name
    """
    mapping = {
        (IORole.INITIATOR, IORole.INITIATOR): "o",
        (IORole.INITIATOR, IORole.RESPONDER): "i",
        (IORole.RESPONDER, IORole.INITIATOR): "i",
        (IORole.RESPONDER, IORole.RESPONDER): "o",
    }
    full_name = f"{bus}_" if bus is not None else ""
    return f"{full_name}{component}_{mapping[role_bus, role_comp]}"


def io_plain_style(bus: str | None, component: str, role_bus: IORole, role_comp: IORole) -> str:
    """
    Style signal names as (<BUS>_)<COMPONENT> for example dma_awaddr and
    dma_awready.

    :param bus:       Name of the bus instance or None if not required
    :param component: Name of component signal withim the bus
    :param role_bus:  Interface role of the entire bus instance
    :param role_comp: Interface role of the component within the bus
    :returns:         Complete string name
    """
    del role_bus
    del role_comp
    if bus is None:
        return component
    else:
        return f"{bus}_{component}"


class BaseIO:
    """
    Wraps a collection of different signals into a single interface that can be
    used by drivers and monitors to interact with the design.

    :param dut:       Pointer to the DUT boundary
    :param name:      Name of the signal - acts as a prefix
    :param role:      Role of this signal on the DUT boundary
    :param init_sigs: Signals driven by the initiator
    :param resp_sigs: Signals driven by the responder
    :param io_style:  Optionally override the default I/O naming style
    """

    DEFAULT_IO_STYLE: Callable[[str | None, str, IORole, IORole], str] = io_prefix_style

    def __init__(
        self,
        dut: HierarchyObject,
        name: str | None,
        role: IORole,
        init_sigs: list[str],
        resp_sigs: list[str],
        io_style: Callable[[str | None, str, IORole, IORole], str] | None = None,
    ) -> None:
        # Sanity checks
        assert role in IORole, f"Role {role} is not recognised"
        assert isinstance(init_sigs, list), "Initiator signals are not a list"
        assert isinstance(resp_sigs, list), "Responder signals are not a list"
        # Hold onto attributes
        self._dut = dut
        self._name = name
        self._role = role
        self._init_sigs = init_sigs[:]
        self._resp_sigs = resp_sigs[:]
        self._defaults = {}
        # If no IO style provided, adopt the default
        io_style = io_style or BaseIO.DEFAULT_IO_STYLE
        # Pickup all initiator and response signals wrapping each inside a
        # SignalWrapper to normalise its behaviour across simulators
        self.__initiators, self.__responders = {}, {}
        for comp in self._init_sigs:
            sig = io_style(self._name, comp, self._role, IORole.INITIATOR)
            if not hasattr(self._dut, sig):
                SimLog("tb").getChild(f"io.{type(self).__name__.lower()}").info(
                    f"{type(self).__name__}: Did not find I/O component {sig} on {dut}"
                )
                continue
            sig_ptr = SignalWrapper(getattr(self._dut, sig))
            self.__initiators[comp] = sig_ptr
            setattr(self, comp, sig_ptr)
        for comp in self._resp_sigs:
            sig = io_style(name, comp, self._role, IORole.RESPONDER)
            if not hasattr(self._dut, sig):
                SimLog("tb").getChild(f"io.{type(self).__name__.lower()}").info(
                    f"{type(self).__name__}: Did not find I/O component {sig} on {dut}"
                )
                continue
            sig_ptr = SignalWrapper(getattr(self._dut, sig))
            self.__responders[comp] = sig_ptr
            setattr(self, comp, sig_ptr)

    @property
    def role(self) -> IORole:
        return self._role

    @property
    def dut(self) -> HierarchyObject:
        return self._dut

    def initialise(self, role: IORole) -> None:
        """Initialise signals according to the active role"""
        for sig in (self.__initiators if role == IORole.INITIATOR else self.__responders).values():
            sig.value = 0

    def set_default(self, comp: str, value: Any) -> None:
        """
        Set the default value to be returned for a signal if it is not available.

        :param comp:  Component name
        :param value: Value to return
        """
        self._defaults[comp] = value

    def has(self, comp: str) -> bool:
        """
        Test whether a particular signal has been resolved inside the interface.

        :param comp: Name of the component
        :returns:    True if exists, False otherwise
        """
        return (comp in self.__initiators) or (comp in self.__responders)

    def get(self, comp: str, default: Any = None) -> Any:
        """
        Get the current value of a particular signal.

        :param comp:    Name of the component
        :param default: Default value if the signal is not resolved
        :returns:       The resolved value, otherwise the default
        """
        item = getattr(self, comp, None)
        if item is None:
            return self._defaults.get(comp, None) if default is None else default
        else:
            raw = int(item.value)
            return (raw == 1) if len(item) == 1 else raw

    def set(self, comp: str, value: Any) -> None:
        """
        Set the value of a particular signal if it exists.

        :param comp:  Name of the component
        :param value: Value to set
        """
        if not self.has(comp):
            return
        getattr(self, comp).value = value

    def width(self, comp: str) -> int:
        """
        Return the width of a particular signal.

        :param comp: Name of the component
        :returns:    The bit width if resolved, else 0
        """
        if not self.has(comp):
            return 0
        sig = getattr(self, comp)
        return max(sig._range) - min(sig._range) + 1



# Copyright 2023, Peter Birch, mailto:peter@lightlogic.co.uk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .bench import BaseBench
from .driver import BaseDriver, DriverEvent
from .io import BaseIO, IORole, io_plain_style, io_prefix_style, io_suffix_style
from .monitor import BaseMonitor, MonitorEvent
from .scoreboard import DrainPolicy, Scoreboard
from .sequence import SeqContext, SeqLock, SeqProxy, randarg, requires, sequence
from .transaction import BaseTransaction

assert all(
    (
        BaseBench,
        BaseDriver,
        IORole,
        BaseIO,
        BaseMonitor,
        BaseTransaction,
        DrainPolicy,
        DriverEvent,
        MonitorEvent,
        Scoreboard,
        SeqContext,
        SeqLock,
        SeqProxy,
        io_prefix_style,
        io_plain_style,
        io_suffix_style,
        randarg,
        requires,
        sequence,
    )
)
